<!DOCTYPE gui SYSTEM "kpartgui.dtd">
<gui name="kdiff3_shell" version="9">
<MenuBar>
  <Menu name="file"><text>&amp;File</text>
    <Action name="file_reload"/>
  </Menu>
  <Menu name="directory"><text>&amp;Directory</text>
    <Action name="dir_start_operation"/>
    <Action name="dir_run_operation_for_current_item"/>
    <Action name="dir_compare_current"/>
    <Action name="dir_rescan"/>
    <!-- <Action name="dir_save_merge_state"/>
    <Action name="dir_load_merge_state"/> -->
    <Action name="dir_fold_all"/>
    <Action name="dir_unfold_all"/>
    <Action name="dir_show_identical_files"/>
    <Action name="dir_show_different_files"/>
    <Action name="dir_show_files_only_in_a"/>
    <Action name="dir_show_files_only_in_b"/>
    <Action name="dir_show_files_only_in_c"/>
    <Action name="dir_choose_a_everywhere"/>
    <Action name="dir_choose_b_everywhere"/>
    <Action name="dir_choose_c_everywhere"/>
    <Action name="dir_autochoose_everywhere"/>
    <Action name="dir_nothing_everywhere"/>
    <Action name="dir_synchronize_directories"/>
    <Action name="dir_choose_newer_files"/>
    <Action name="dir_compare_explicitly_selected_files"/>
    <Action name="dir_merge_explicitly_selected_files"/>
    <Menu name="dir_current_merge_menu"><text>Current Item Merge Operation</text>
       <Action name="dir_current_do_nothing"/>
       <Action name="dir_current_choose_a"/>
       <Action name="dir_current_choose_b"/>
       <Action name="dir_current_choose_c"/>
       <Action name="dir_current_merge"/>
       <Action name="dir_current_delete"/>
    </Menu>
    <Menu name="dir_current_sync_menu"><text>Current Item Sync Operation</text>
       <Action name="dir_current_sync_do_nothing"/>
       <Action name="dir_current_sync_copy_a_to_b"/>
       <Action name="dir_current_sync_copy_b_to_a"/>
       <Action name="dir_current_sync_delete_a"/>
       <Action name="dir_current_sync_delete_b"/>
       <Action name="dir_current_sync_delete_a_and_b"/>
       <Action name="dir_current_sync_merge_to_a"/>
       <Action name="dir_current_sync_merge_to_b"/>
       <Action name="dir_current_sync_merge_to_a_and_b"/>
    </Menu>
  </Menu>
  <Menu name="movement"><text>M&amp;ovement</text>
    <Action name="go_current"/>
    <Action name="go_top"/>
    <Action name="go_bottom"/>
    <Action name="go_prev_delta"/>
    <Action name="go_next_delta"/>
    <Action name="go_prev_conflict"/>
    <Action name="go_next_conflict"/>
    <Action name="go_prev_unsolved_conflict"/>
    <Action name="go_next_unsolved_conflict"/>
  </Menu>
  <Menu name="diff"><text>D&amp;iffview</text>
    <Action name="diff_showlinenumbers"/>
    <Action name="diff_show_whitespace_characters"/>
    <Action name="diff_show_whitespace"/>
    <Action name="diff_overview_normal"/>
    <Action name="diff_overview_ab"/>
    <Action name="diff_overview_ac"/>
    <Action name="diff_overview_bc"/>
    <Action name="diff_wordwrap"/>
    <Action name="diff_add_manual_diff_help"/>
    <Action name="diff_clear_manual_diff_help_list"/>
  </Menu>  
  <Menu name="merge"><text>M&amp;erge</text>
    <Action name="merge_current"/>
    <Action name="merge_choose_a"/>
    <Action name="merge_choose_b"/>
    <Action name="merge_choose_c"/>
    <Action name="merge_autoadvance"/>
    <Action name="merge_choose_a_everywhere"/>
    <Action name="merge_choose_b_everywhere"/>
    <Action name="merge_choose_c_everywhere"/>
    <Action name="merge_choose_a_for_unsolved_conflicts"/>
    <Action name="merge_choose_b_for_unsolved_conflicts"/>
    <Action name="merge_choose_c_for_unsolved_conflicts"/>
    <Action name="merge_choose_a_for_unsolved_whitespace_conflicts"/>
    <Action name="merge_choose_b_for_unsolved_whitespace_conflicts"/>
    <Action name="merge_choose_c_for_unsolved_whitespace_conflicts"/>
    <Action name="merge_autosolve"/>
    <Action name="merge_autounsolve"/>
    <Action name="merge_splitdiff"/>
    <Action name="merge_joindiffs"/>
    <Action name="merge_regexp_automerge"/>
    <Action name="merge_versioncontrol_history"/>
  </Menu>
  <Menu name="window"><text>&amp;Window</text>
    <Action name="win_focus_prev"/>
    <Action name="win_focus_next"/>
    <Action name="win_show_a"/>
    <Action name="win_show_b"/>
    <Action name="win_show_c"/>
    <Action name="win_dir_show_both"/>
    <Action name="win_dir_view_toggle"/>
    <Action name="win_toggle_split_orientation"/>
  </Menu>
</MenuBar>
<ToolBar name="mainToolBar" iconText="icononly" fullWidth="true"><text>Main Toolbar</text>
  <Action name="merge_current"/>
  <Action name="go_current"/>
  <Action name="go_top"/>
  <Action name="go_bottom"/>
  <Action name="go_prev_delta"/>
  <Action name="go_next_delta"/>
  <Action name="go_prev_conflict"/>
  <Action name="go_next_conflict"/>
  <Action name="go_prev_unsolved_conflict"/>
  <Action name="go_next_unsolved_conflict"/>
  <Action name="merge_choose_a"/>
  <Action name="merge_choose_b"/>
  <Action name="merge_choose_c"/>
  <Action name="merge_autoadvance"/>
  <Action name="diff_show_whitespace"/>
  <Action name="diff_show_whitespace_characters"/>
  <Action name="diff_showlinenumbers"/>
  <Action name="dir_show_identical_files"/>
  <Action name="dir_show_files_only_in_a"/>
  <Action name="dir_show_files_only_in_b"/>
</ToolBar>
</gui>
